from typing import Union, Literal, Optional, Any, Dict

from bolinette import types, core
from bolinette.exceptions import InitError


class Reference:
    def __init__(self, model: 'core.Model', column: 'core.models.Column', models: Dict[str, 'core.Model'],
                 model_name: str, column_name: str):
        self.model = model
        self.column = column
        if model_name not in models:
            raise InitError(f'{model.__blnt__.name}.{column.name}: unknown "{model_name}" model')
        self.target_model = models[model_name]
        target_cols = dict(self.target_model.__props__.get_columns())
        if column_name not in target_cols:
            raise InitError(f'{model.__blnt__.name}.{column.name}: no "{column_name}" column in "{model_name}" model')
        self.target_column = target_cols[column_name]

    @property
    def model_name(self):
        return self.model.__blnt__.name

    @property
    def column_name(self):
        return self.column.name

    @property
    def target_model_name(self):
        return self.target_model.__blnt__.name

    @property
    def target_column_name(self):
        return self.target_column.name

    @property
    def target_path(self):
        return f'{self.target_model_name}.{self.target_column_name}'

    def __repr__(self):
        return f'<Reference {self.model_name}.{self.column_name} -> {self.target_path}>'


class Column:
    def __init__(self, name: str, model: 'core.Model', data_type: 'types.db.DataType', reference: Optional[Reference],
                 primary_key: bool, nullable: bool, unique: bool, model_id: bool, default: Optional[Any]):
        self.name = name
        self.type = data_type
        self.model = model
        self.reference = reference
        self.primary_key = primary_key
        self.nullable = nullable
        self.unique = unique
        self.model_id = model_id
        self.default = default

    @property
    def model_name(self):
        return self.model.__blnt__.name

    def __repr__(self):
        s = f'<Column {self.model_name}.{self.name}: {repr(self.type)}'
        if self.reference is not None:
            s += f' -> {self.reference.target_path}'
        return s + '>'


class ColumnList:
    def __init__(self, name: str, model: 'core.Model', origin: 'core.Model'):
        self.name = name
        self.model = model
        self.origin = origin


class Backref:
    def __init__(self, model: 'core.Model', relationship: 'core.models.Relationship',
                 key: str, lazy: bool):
        self.model = model
        self.relationship = relationship
        self.key = key
        self.lazy = lazy

    def __repr__(self):
        return f'<Backref <- {self.key}' + (' (lazy)' if self.lazy else '') + '>'


class Relationship:
    def __init__(self, name: str, model: 'core.Model', models: Dict[str, 'core.Model'],
                 model_name: str, backref: Optional[Backref], foreign_key: Optional[Column],
                 remote_side: Optional[Column], lazy: Union[bool, Literal['subquery']], secondary: Optional[str]):
        self.name = name
        self.model = model
        if model_name not in models:
            raise InitError(f'{model.__blnt__.name}.{name}: unknown "{model_name}" model')
        self.target_model = models[model_name]
        self.foreign_key = foreign_key
        self.remote_side = remote_side
        if secondary is not None and secondary not in models:
            raise InitError(f'{model.__blnt__.name}.{name}: unknown "{secondary}" model')
        self.secondary = models[secondary] if secondary is not None else None
        self.backref = backref
        self.lazy = lazy

    @property
    def model_name(self):
        return self.model.__blnt__.name

    @property
    def target_model_name(self):
        return self.target_model.__blnt__.name

    def __repr__(self):
        return (f'<Relationship {self.model_name}.{self.name} -> {self.target_model_name}'
                + ('  (lazy)' if self.lazy else '') + '>')



from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from bolinette.blnt.database.engines import DatabaseEngine


class RelationalDatabase(DatabaseEngine):
    def __init__(self, uri):
        super().__init__(relational=True)
        self._engine = create_engine(uri, echo=False)
        self._base = declarative_base()
        self._Session = sessionmaker(bind=self._engine)
        self._session = self._Session()

    @property
    def base(self):
        return self._base

    @property
    def session(self):
        return self._session

    async def open_transaction(self):
        pass

    async def close_transaction(self):
        self._session.commit()

    async def rollback_transaction(self):
        self._session.rollback()

    async def create_all(self):
        self._base.metadata.create_all(self._engine)

    async def drop_all(self):
        self._base.metadata.drop_all(self._engine)



import asyncio
import sys
from datetime import datetime
from typing import Dict, Any, List

from aiohttp import web as aio_web

from bolinette import blnt, BolinetteExtension, Extensions, Console
from bolinette.blnt.commands import Parser
from bolinette.exceptions import InitError
from bolinette.utils import paths


class Bolinette:
    def __init__(self, *, extensions: List[BolinetteExtension] = None,
                 profile: str = None, overrides: Dict[str, Any] = None):
        self._start_time = datetime.utcnow()
        self._init = False
        self._init_ext = False
        self.app = None
        try:
            self.context = blnt.BolinetteContext(paths.dirname(__file__), extensions=extensions,
                                                 profile=profile, overrides=overrides)
            self.context.use_extension(Extensions.ALL)
        except InitError as init_error:
            Console().error(f'Error raised during Bolinette init phase\n{str(init_error)}')
            exit(1)
        self.console = Console(debug=self.context.env.debug)

    def _run_init_functions(self):
        index = 0
        for func in blnt.cache.init_funcs:
            if not self.context.has_extension(func.extension):
                continue
            loop = asyncio.get_event_loop()
            if index == 0:
                self.console.debug('Running Bolinette init functions: ', end='')
            else:
                self.console.debug(', ', end='')
            self.console.debug(str(func), end='')
            loop.run_until_complete(func(self.context))
            index += 1
        if index > 0:
            self.console.debug('')

    def init(self, *, force=False):
        if self._init and not force:
            return
        self._run_init_functions()
        self._init = True

    def init_extensions(self, *, force=False):
        if self._init_ext and not force:
            return
        self.app = None
        if self.context.has_extension(Extensions.WEB):
            if self.context.env['build_docs']:
                self.context.docs.build()
            self.context.docs.setup()
            self._init_web()
        if self.context.has_extension(Extensions.SOCKETS):
            self._init_sockets()
        self._init_ext = True

    def _init_web(self):
        if self.app is None:
            self.app = aio_web.Application()
            self.app['blnt'] = self.context
        self.context.init_web(self.app)

    def _init_sockets(self):
        if self.app is None:
            self.app = aio_web.Application()
            self.app['blnt'] = self.context
        self.context.init_sockets(self.app)

    def start_server(self, *, host: str = None, port: int = None):
        if not self.context.has_extension((Extensions.WEB, Extensions.SOCKETS)):
            self.context.logger.error(f'The web or sockets extensions must be activated to start the aiohttp server!')
            sys.exit(1)
        self.init()
        self.init_extensions()
        self.console.debug(f'Startup took {int((datetime.utcnow() - self._start_time).microseconds / 1000)}ms')
        self.context.logger.info(f"Starting Bolinette with '{self.context.env['profile']}' environment profile")
        aio_web.run_app(self.app,
                        host=host or self.context.env['host'],
                        port=port or self.context.env['port'],
                        access_log=self.context.logger)
        self.context.logger.info(f"Bolinette stopped gracefully")

    def use(self, *extensions) -> 'Bolinette':
        self.context.clear_extensions()
        for ext in extensions:
            self.context.use_extension(ext)
        return self

    def exec_cmd_args(self):
        self.init()
        self.init_extensions()
        parser = Parser(self, blnt.cache.commands)
        parser.run()



from bolinette import types, core, blnt, mapping
from bolinette.decorators import model


@model('users_roles', join_table=True)
class UsersRoles(core.Model):
    user_id = types.defs.Column(types.db.Integer, reference=types.defs.Reference('user', 'id'), primary_key=True)
    role_id = types.defs.Column(types.db.Integer, reference=types.defs.Reference('role', 'id'), primary_key=True)


@model('user')
class User(core.Model):
    id = types.defs.Column(types.db.Integer, primary_key=True)
    username = types.defs.Column(types.db.String, unique=True, nullable=False, model_id=True)
    password = types.defs.Column(types.db.Password, nullable=False)
    email = types.defs.Column(types.db.Email, unique=blnt.init['user_email_required'],
                              nullable=(not blnt.init['user_email_required']))

    roles = types.defs.Relationship('role', secondary='users_roles', lazy='subquery',
                                    backref=types.defs.Backref('users', lazy=True))

    picture_id = types.defs.Column(types.db.Integer, reference=types.defs.Reference('file', 'id'))
    profile_picture = types.defs.Relationship('file', foreign_key=picture_id, lazy=False)

    timezone = types.defs.Column(types.db.String)

    def payloads(self):
        yield 'register', [
            mapping.Column(self.username, required=True),
            mapping.Column(self.email, required=True),
            mapping.Column(self.password, required=True),
            mapping.Column(self.timezone)
        ]
        yield 'admin_register', [
            mapping.Column(self.username, required=True),
            mapping.Column(self.email, required=True),
            mapping.Field(types.db.Boolean, name='send_mail', required=True)
        ]
        yield 'login', [
            mapping.Column(self.username, required=True),
            mapping.Column(self.password, required=True)
        ]

    def responses(self):
        default = [
            mapping.Column(self.username),
            mapping.Reference(self.profile_picture, 'minimal')
        ]
        yield default
        yield 'private', default + [
            mapping.Column(self.email),
            mapping.List(mapping.Definition('role'), key='roles'),
            mapping.Column(self.timezone)
        ]



import datetime
import random
import string
from types import SimpleNamespace
from typing import Dict, Any

from bolinette import blnt, types


class Mocked:
    def __init__(self, name, context: 'blnt.BolinetteContext'):
        self.context = context
        self.name = name
        self.model = self.context.model(self.name)
        self.database = self.context.db[self.model.__blnt__.database]
        self._fields = {}

    def __getitem__(self, key):
        return self._fields[key]

    def __setitem__(self, key, value):
        self._fields[key] = value

    def __contains__(self, key):
        return key in self._fields

    def __repr__(self):
        return repr(self._fields)

    @staticmethod
    async def insert_entity(context: 'blnt.BolinetteContext', name: str, params: Dict[str, Any]):
        mocked = Mocked(name, context)
        for key, value in params.items():
            mocked[key] = value
        return await mocked.insert()

    @property
    def _to_object(self):
        obj = SimpleNamespace()
        for key, value in self._fields.items():
            setattr(obj, key, value)
        return obj

    async def insert(self):
        return await self.context.repo(self.name).create(self._fields)

    def to_response(self, key='default') -> dict:
        definition = self.context.mapper.response(self.name, key)
        return self.context.mapper.marshall(
            definition,
            self._to_object if self.database.relational else self._fields,
            use_foreign_key=True
        )

    def to_payload(self, key='default') -> dict:
        definition = self.context.mapper.payload(self.name, key)
        return self.context.mapper.marshall(
            definition,
            self._to_object if self.database.relational else self._fields,
            use_foreign_key=True
        )


class Mock:
    def __init__(self, context: 'blnt.BolinetteContext'):
        self._id = None
        self.context = context

    def _random_lower(self, rng, length):
        return ''.join(rng.choices(string.ascii_lowercase, k=length))

    def _random_symbols(self, rng, length):
        return ''.join(rng.choices(string.punctuation, k=length))

    def _random_int(self, rng, a, b):
        return rng.randint(a, b)

    def _random_decimal(self, rng, a, b):
        return rng.uniform(a, b)

    def _random_date(self, rng, start_date, end_date):
        time_between_dates = end_date - start_date
        days_between_dates = time_between_dates.days
        random_number_of_days = rng.randrange(days_between_dates)
        return start_date + datetime.timedelta(days=random_number_of_days)

    def __call__(self, m_id, model_name, *, post_mock_fn=None):
        self._id = m_id
        rng = random.Random(hash(f'{model_name}.{m_id}'))
        mocked = Mocked(model_name, self.context)
        model = self.context.model(model_name)
        columns = model.__props__.get_columns()
        for _, column in columns:
            if column.primary_key:
                continue
            col_type = column.type
            if col_type == types.db.String:
                mocked[column.name] = self._random_lower(rng, 15)
            if col_type == types.db.Email:
                mocked[column.name] = f'{self._random_lower(rng, 10)}@{self._random_lower(rng, 5)}.com'
            if col_type == types.db.Password:
                mocked[column.name] = (self._random_lower(rng, 10) + str(self._random_int(rng, 1, 100))
                                       + self._random_symbols(rng, 1))
            if col_type == types.db.Integer:
                mocked[column.name] = self._random_int(rng, 1, 100)
            if col_type == types.db.Float:
                mocked[column.name] = self._random_int(rng, 1, 100)
            if col_type == types.db.Date:
                mocked[column.name] = self._random_date(rng, datetime.datetime(1900, 1, 1),
                                                        datetime.datetime(2000, 1, 1))
        back_refs = model.__props__.get_back_refs()
        for att_name, _ in back_refs:
            mocked[att_name] = []
        if post_mock_fn and callable(post_mock_fn):
            post_mock_fn(mocked)
        return mocked



import random
import string

from bolinette.utils import paths

from bolinette import core
from bolinette.decorators import service


@service('file')
class FileService(core.Service):
    async def _generate_key(self):
        key = None
        while key is None or len(await self.get_by('key', key)):
            key = ''.join(random.choices(string.ascii_lowercase + string.digits, k=32))
        return key

    async def write_file(self, content, key):
        path = self.context.instance_path('uploads')
        if not paths.exists(path):
            paths.mkdir(path)
        with open(paths.join(path, key), 'wb') as f:
            f.write(content)

    async def delete_file(self, key):
        paths.rm(self.context.instance_path('uploads', key))

    async def delete(self, entity):
        ent = await super().delete(entity)
        await self.delete_file(entity.key)
        return ent

    async def save_file(self, request_file):
        params = {
            'key': await self._generate_key(),
            'name': request_file.filename,
            'mime': request_file.content_type
        }
        await self.write_file(request_file.file.read(), params['key'])
        return await self.create(params)

    async def file_sender(self, key):
        with open(self.context.instance_path('uploads', key), 'rb') as f:
            chunk = f.read(2 ** 16)
            while chunk:
                yield chunk
                chunk = f.read(2 ** 16)



from typing import List, Union, Tuple, Dict, Iterator, Literal

from bolinette import core, blnt
from bolinette.blnt.database.engines import DatabaseEngine
from bolinette.exceptions import InitError

MappingPyTyping = List[Union['types.mapping.Column', 'types.mapping.Field',
                             'types.mapping.List', 'types.mapping.Definition']]
MappingListPyTyping = Union[MappingPyTyping, Tuple[str, MappingPyTyping]]


class Model:
    __blnt__: 'ModelMetadata' = None

    def __init__(self, database: 'DatabaseEngine'):
        self.__props__ = ModelProps(self, database)

    @classmethod
    def payloads(cls) -> MappingListPyTyping:
        pass

    @classmethod
    def responses(cls) -> MappingListPyTyping:
        pass

    def get_mixin(self, name: str):
        return self.__props__.mixins[name]

    def __repr__(self):
        return f'<Model {self.__blnt__.name}>'


class ModelMetadata:
    def __init__(self, name: str, database: str, relational: bool, join: bool, mixins: List[str],
                 merge_defs: Literal['ignore', 'append', 'overwrite']):
        self.name = name
        self.database = database
        self.relational = relational
        self.join = join
        self.mixins = mixins
        self.merge_defs = merge_defs


class ModelProps(blnt.Properties):
    def __init__(self, model: Model, database: 'DatabaseEngine'):
        super().__init__(model)
        self.model = model
        self.database = database
        self.mixins: Dict[str, core.Mixin] = {}
        # self.model_id: 'core.models.Column' = self._set_model_id()

    @property
    def model_id(self):
        attrs = filter(lambda col: col[1].model_id, self.get_columns())
        model_id = next(attrs)
        return model_id[1]

    def _set_model_id(self):
        model_id = None
        if self.model.__blnt__.join:
            return model_id
        for col_name, column in self.get_columns():
            if column.model_id and model_id is not None:
                raise InitError(f'Model "{self.model.__blnt__.name}" already has a model id')
            if column.model_id:
                model_id = column
        if model_id is None:
            raise InitError(f'Model "{self.model.__blnt__.name}" has no column marked as model id')
        if not model_id.unique and not model_id.primary_key:
            raise InitError(f'Model "{self.model.__blnt__.name}"\'s model id should be either '
                            'a unique column or a primary key')
        return model_id

    def get_columns(self) -> Iterator[Tuple[str, 'core.models.Column']]:
        return self._get_attributes_of_type(self.parent, core.models.Column)

    def get_relationships(self) -> Iterator[Tuple[str, 'core.models.Relationship']]:
        return self._get_attributes_of_type(self.parent, core.models.Relationship)

    def get_properties(self) -> Iterator[Tuple[str, 'ModelProperty']]:
        return self._get_cls_attributes_of_type(type(self.parent), ModelProperty)

    def get_back_refs(self) -> Iterator[Tuple[str, 'core.models.ColumnList']]:
        return self._get_attributes_of_type(self.parent, core.models.ColumnList)


class ModelProperty:
    def __init__(self, name, function):
        self.name = name
        self.function = function

    def __repr__(self):
        return f'<ModelProperty {self.name}>'



from os import PathLike
from typing import Union, List

from aiohttp.web_response import Response as AioResponse
from jinja2 import TemplateNotFound

from bolinette.utils import files, paths

from bolinette import blnt, exceptions


class Cookie:
    def __init__(self, name, value, *, path, http_only=True, expires=None, delete=False):
        self.name = name
        self.value = value
        self.expires = expires
        self.path = path
        self.http_only = http_only
        self.delete = delete


class APIResponse:
    def __init__(self, content, code):
        self.content = content
        self.code = code
        self.cookies = []


class Response:
    def __init__(self, context: 'blnt.BolinetteContext'):
        self._exceptions = {
            exceptions.UnauthorizedError: self.unauthorized,
            exceptions.BadRequestError: self.bad_request,
            exceptions.UnprocessableEntityError: self.unprocessable_entity,
            exceptions.ConflictError: self.conflict,
            exceptions.NotFoundError: self.not_found,
            exceptions.ForbiddenError: self.forbidden,
            exceptions.APIError: self.internal_server_error
        }
        self.context = context

    @staticmethod
    def _build_message(code, status, messages=None, data=None):
        if messages is None:
            messages = []
        if not isinstance(messages, list):
            messages = [messages]
        body = {
            'code': code,
            'status': status,
            'messages': messages
        }
        if data is not None:
            body['data'] = data
        return APIResponse(body, code)

    def ok(self, *, messages: Union[str, List[str]] = None, data=None):
        return self._build_message(200, 'OK', messages, data)

    def created(self, *, messages: Union[str, List[str]] = None, data=None):
        return self._build_message(201, 'CREATED', messages, data)

    def bad_request(self, *, messages: Union[str, List[str]] = None, data=None):
        return self._build_message(400, 'BAD REQUEST', messages, data)

    def unauthorized(self, *, messages: Union[str, List[str]] = None, data=None):
        return self._build_message(401, 'UNAUTHORIZED', messages, data)

    def forbidden(self, *, messages: Union[str, List[str]] = None, data=None):
        return self._build_message(403, 'FORBIDDEN', messages, data)

    def not_found(self, *, messages: Union[str, List[str]] = None, data=None):
        return self._build_message(404, 'NOT FOUND', messages, data)

    def method_not_allowed(self, *, messages: Union[str, List[str]] = None, data=None):
        return self._build_message(405, 'METHOD NOT ALLOWED', messages, data)

    def conflict(self, *, messages: Union[str, List[str]] = None, data=None):
        return self._build_message(409, 'CONFLICT', messages, data)

    def unprocessable_entity(self, *, messages: Union[str, List[str]] = None, data=None):
        return self._build_message(422, 'UNPROCESSABLE ENTITY', messages, data)

    def internal_server_error(self, *, messages: Union[str, List[str]] = None, data=None):
        return self._build_message(500, 'INTERNAL SERVER ERROR', messages, data)

    def render_template(self, name: PathLike, params: dict = None, workdir: PathLike = None, *, status=200):
        if workdir is None:
            workdir = self.context.templates_path()
        if params is None:
            params = {}
        for key, value in self.context.manifest.items():
            if key not in params:
                params[key] = value
        try:
            content = files.render_template(workdir, name, params)
        except TemplateNotFound:
            error_404_wd = self.context.internal_files_path('templates')
            error_404 = paths.join('errors', '404.html.jinja2')
            content = files.render_template(error_404_wd, error_404, params)
            status = 404
        return AioResponse(body=content, status=status, content_type='text/html')

    def from_exception(self, exception: Union[exceptions.APIError, exceptions.APIErrors]):
        messages = []
        if isinstance(exception, exceptions.APIError):
            messages = [exception.message]
        elif isinstance(exception, exceptions.APIErrors):
            messages = [error.message for error in exception.errors]
            exception = exception.errors[0]
        for except_cls in self._exceptions:
            if isinstance(exception, except_cls):
                return self._exceptions[except_cls](messages=messages)



from bolinette import web
from bolinette.decorators import middleware
from bolinette.exceptions import ForbiddenError, UnauthorizedError


@middleware('auth', priority=20)
class AuthMiddleware(web.InternalMiddleware):
    def define_options(self):
        return {
            'optional': self.params.bool(),
            'fresh': self.params.bool(),
            'roles': self.params.list(self.params.string())
        }

    async def handle(self, request, params, next_func):
        identity = self.context.jwt.verify(request, optional=self.options['optional'],
                                           fresh=self.options['fresh'])
        current_user = None
        if identity is not None:
            current_user = await self.context.service('user').get_by_username(identity)
        if len(self.options['roles']) > 0:
            if current_user is None:
                raise UnauthorizedError('user.unauthorized')
            roles = self.options['roles']
            if isinstance(roles, str):
                roles = [roles]
            if current_user is None:
                raise ForbiddenError(f'user.forbidden:{",".join(roles)}')
            user_roles = set(map(lambda r: r.name, current_user.roles))
            if 'root' not in user_roles and not len(user_roles.intersection(set(roles))):
                raise ForbiddenError(f'user.forbidden:{",".join(roles)}')
        params['current_user'] = current_user
        return await next_func(request, params)



from bson import ObjectId
from pymongo import ASCENDING, DESCENDING
from pymongo.collection import Collection
from pymongo.cursor import Cursor

from bolinette import blnt, core
from bolinette.blnt.database.engines import CollectionDatabase
from bolinette.blnt.database.queries import BaseQueryBuilder, BaseQuery


class CollectionQueryBuilder(BaseQueryBuilder):
    def __init__(self, model: 'core.Model', context: 'blnt.BolinetteContext'):
        super().__init__(model, context)
        self._name = model.__blnt__.name
        self._database: 'CollectionDatabase' = context.db[model.__blnt__.database]
        self._collection = self._database.db[self._name]

    def query(self) -> 'BaseQuery':
        return CollectionQuery(self._collection)

    async def insert_entity(self, values):
        result = self._collection.insert_one(values)
        return self._collection.find_one({'_id': ObjectId(result.inserted_id)})

    async def update_entity(self, entity):
        values = dict(entity)
        _id = values.pop('_id')
        self._collection.update_one({'_id': _id}, {'$set': values})
        return self._collection.find_one({'_id': ObjectId(_id)})

    async def delete_entity(self, entity):
        self._collection.delete_one({'_id': entity['_id']})
        return entity


class CollectionQuery(BaseQuery):
    def __init__(self, collection: Collection):
        super().__init__()
        self._collection = collection

    def _clone(self):
        query = CollectionQuery(self._collection)
        self._base_clone(query)
        return query

    async def all(self):
        return self._apply_params(self._collection.find(self._build_filters()))

    async def first(self):
        return self._collection.find_one(self._build_filters())

    async def get_by_id(self, identifier):
        return self._collection.find_one({'_id': ObjectId(identifier)})

    async def count(self):
        return self._collection.count_documents(self._build_filters())

    def _apply_params(self, cursor: Cursor):
        cursor.skip(self._offset)
        if self._limit is not None:
            cursor.limit(self._limit)
        if len(self._order_by) > 0:
            order_by = [(column, DESCENDING if desc else ASCENDING) for column, desc in self._order_by]
            cursor.sort(order_by)
        return cursor

    def _build_filters(self):
        params = {}
        if len(self._filters_by) > 0:
            for key in self._filters_by:
                if key == '_id':
                    params[key] = ObjectId(self._filters_by[key])
                else:
                    params[key] = self._filters_by[key]
        return params



from bolinette import core
from bolinette.decorators import service


@service('role')
class RoleService(core.Service):
    async def get_by_name(self, name):
        return await self.get_first_by('name', name)



import itertools
from typing import Optional, Type, Iterator, Tuple

from bolinette import utils


class Properties:
    def __init__(self, parent):
        self.parent = parent

    @staticmethod
    def _get_cls_attributes_of_type(obj: Type, attr_type):
        parent_attrs = (Properties._get_cls_attributes_of_type(parent, attr_type) for parent in obj.__bases__)
        return itertools.chain(
            *parent_attrs,
            ((name, attribute) for name, attribute in vars(obj).items() if isinstance(attribute, attr_type))
        )

    @staticmethod
    def _get_attributes_of_type(obj, attr_type):
        return ((name, attribute)
                for name, attribute in vars(obj).items()
                if isinstance(attribute, attr_type))

    def get_proxies(self, of_type: Optional[Type] = None) -> Iterator[Tuple[str, 'utils.InitProxy']]:
        proxies = self._get_cls_attributes_of_type(type(self.parent), utils.InitProxy)
        if of_type is not None:
            return filter(lambda p: p[1].of_type(of_type), proxies)
        return proxies



from bolinette import web
from bolinette.decorators import controller, get
from bolinette.defaults.services import TimezoneService


@controller('tz', '/tz')
class TimezoneController(web.Controller):
    @property
    def tz_service(self) -> TimezoneService:
        return self.context.service('tz')

    @get('')
    async def all_timezones(self):
        """
        Gets all available IANA timezones
        """
        return self.response.ok(data=await self.tz_service.get_all())



from bolinette import web
from bolinette.decorators import controller


@controller('role', '/role')
class RoleController(web.Controller):
    def default_routes(self):
        return [
            self.defaults.get_all(middlewares=['auth|roles=admin']),
            self.defaults.get_one('complete', key='name', middlewares=['auth|roles=admin'])
        ]



from typing import List, Dict, Any

from bolinette import blnt
from bolinette.blnt.objects import PaginationParams, OrderByParams
from bolinette.exceptions import EntityNotFoundError
from bolinette.utils.functions import async_invoke


class Service:
    __blnt__: 'ServiceMetadata' = None

    def __init__(self, context: 'blnt.BolinetteContext'):
        self.__props__ = ServiceProps(self)
        self.context = context
        self.repo = context.repo(self.__blnt__.model_name)

    def __repr__(self):
        return f'<Service {self.__blnt__.name}>'

    async def get(self, identifier, *, safe=False):
        entity = await self.repo.get(identifier)
        if entity is None and not safe:
            raise EntityNotFoundError(model=self.__blnt__.name, key='id', value=identifier)
        return entity

    async def get_by(self, key: str, value):
        return await self.repo.get_by(key, value)

    async def get_first_by(self, key: str, value, *, safe=False):
        entity = await self.repo.get_first_by(key, value)
        if entity is None and not safe:
            raise EntityNotFoundError(model=self.__blnt__.name, key=key, value=value)
        return entity

    async def get_all(self, *, pagination: PaginationParams = None, order_by: List[OrderByParams] = None):
        return await self.repo.get_all(pagination, order_by)

    async def create(self, values: Dict[str, Any], **kwargs):
        await self.__props__.call_mixin_methods('create', self, values=values, **kwargs)
        return await self.repo.create(values)

    async def update(self, entity, values: Dict[str, Any], **kwargs):
        await self.__props__.call_mixin_methods('update', self, entity=entity, values=values, **kwargs)
        return await self.repo.update(entity, values)

    async def patch(self, entity, values: Dict[str, Any], **kwargs):
        await self.__props__.call_mixin_methods('patch', self, entity=entity, values=values, **kwargs)
        return await self.repo.patch(entity, values)

    async def delete(self, entity, **kwargs):
        await self.__props__.call_mixin_methods('delete', self, entity=entity, **kwargs)
        return await self.repo.delete(entity)


class SimpleService:
    __blnt__: 'ServiceMetadata' = None

    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context

    def __repr__(self):
        return f'<Service {self.__blnt__.name}>'


class ServiceMetadata:
    def __init__(self, name: str, model_name: str):
        self.name = name
        self.model_name = model_name


class ServiceProps(blnt.Properties):
    async def call_mixin_methods(self, method_name: str, *args, **kwargs):
        for _, mixin in self.parent.repo.model.__props__.mixins.items():
            if method_name in mixin:
                await async_invoke(mixin[method_name].func, *args, **kwargs)



from typing import Optional, Any, Literal, Union

from bolinette import types, core
from bolinette.utils import InitProxy


class Reference:
    def __new__(cls, model_name: str, column_name: str) -> InitProxy['core.models.Reference']:
        return InitProxy(core.models.Reference, model_name=model_name, column_name=column_name)


class Column:
    def __new__(cls, data_type: 'types.db.DataType', *,
                reference: Optional['core.models.Reference'] = None,
                primary_key: bool = False,
                nullable: bool = True,
                unique: bool = False,
                model_id: bool = False,
                default: Optional[Any] = None) -> InitProxy['core.models.Column']:
        return InitProxy(core.models.Column, data_type=data_type, reference=reference, primary_key=primary_key,
                         nullable=nullable, unique=unique, model_id=model_id, default=default)


class Backref:
    def __new__(cls, key: str, *, lazy: bool = True) -> InitProxy['core.models.Backref']:
        return InitProxy(core.models.Backref, key=key, lazy=lazy)


class Relationship:
    def __new__(cls, model_name: str, *,
                backref: 'core.models.Backref' = None,
                foreign_key: 'core.models.Column' = None,
                lazy: Union[bool, Literal['subquery']] = False,
                secondary: str = None,
                remote_side: 'core.models.Column' = None) -> InitProxy['core.models.Relationship']:
        return InitProxy(core.models.Relationship, model_name=model_name, backref=backref, foreign_key=foreign_key,
                         lazy=lazy, secondary=secondary, remote_side=remote_side)



import inspect
from typing import Callable, Awaitable, Any

from bolinette.exceptions import InternalError


def _parse_params(function, *args, **kwargs):
    cur_arg = 0
    arg_cnt = len(args)
    out_args = []
    out_kwargs = {}
    for key, param in inspect.signature(function).parameters.items():
        if param.kind == param.POSITIONAL_ONLY:
            if cur_arg < arg_cnt:
                out_args.append(args[cur_arg])
                cur_arg += 1
            else:
                out_args.append(param.default if not param.empty else None)
        elif param.kind == param.KEYWORD_ONLY:
            out_kwargs[key] = kwargs.pop(key, param.default if not param.empty else None)
        elif param.kind == param.POSITIONAL_OR_KEYWORD:
            if cur_arg < arg_cnt:
                out_args.append(args[cur_arg])
                cur_arg += 1
            else:
                out_args.append(kwargs.pop(key, param.default if not param.empty else None))
        elif param.kind == param.VAR_POSITIONAL:
            while cur_arg < arg_cnt:
                out_args.append(args[cur_arg])
                cur_arg += 1
        elif param.kind == param.VAR_KEYWORD:
            for p_name, p_value in kwargs.items():
                out_kwargs[p_name] = p_value
    return out_args, out_kwargs


async def async_invoke(function: Callable[[Any], Awaitable[Any]], *args, **kwargs):
    if inspect.iscoroutinefunction(function):
        args, kwargs = _parse_params(function, *args, **kwargs)
        return await function(*args, **kwargs)
    raise InternalError(f'internal.not_async_function:{function.__name__}')


def invoke(function: Callable[[Any], Any], *args, **kwargs):
    if inspect.isfunction(function):
        args, kwargs = _parse_params(function, *args, **kwargs)
        return function(*args, **kwargs)
    raise InternalError(f'internal.not_function:{function.__name__}')


def getattr_(entity, key, default):
    if isinstance(entity, dict):
        return entity.get(key, default)
    return getattr(entity, key, default)


def hasattr_(entity, key):
    if isinstance(entity, dict):
        return key in entity
    return hasattr(entity, key)


def setattr_(entity, key, value):
    if isinstance(entity, dict):
        entity[key] = value
    else:
        setattr(entity, key, value)



from bolinette import types, core, mapping
from bolinette.decorators import model


@model('file')
class File(core.Model):
    id = types.defs.Column(types.db.Integer, primary_key=True)
    key = types.defs.Column(types.db.String, nullable=False, unique=True, model_id=True)
    name = types.defs.Column(types.db.String, nullable=False)
    mime = types.defs.Column(types.db.String, nullable=False)

    def responses(self):
        yield [
            mapping.Column(self.key),
            mapping.Column(self.name),
            mapping.Column(self.mime)
        ]
        yield 'minimal', [
            mapping.Column(self.key)
        ]



import bolinette.defaults.models
import bolinette.defaults.mixins
import bolinette.defaults.services
import bolinette.defaults.middlewares
import bolinette.defaults.controllers
import bolinette.defaults.topics



import asyncio
import inspect
import sys
from argparse import ArgumentParser
from typing import Dict

import bolinette
from bolinette import Console
from bolinette.blnt.commands import Command, Argument
from bolinette.utils.functions import async_invoke, invoke


class Parser:
    def __init__(self, blnt: 'bolinette.Bolinette', commands: Dict[str, Command]):
        self.blnt = blnt
        self.commands = commands
        self._factories = {
            'argument': self._create_argument,
            'option': self._create_option,
            'flag': self._create_flag
        }

    def run(self):
        parser = ArgumentParser(description='Bolinette Web Framework')
        sub_parsers = parser.add_subparsers()
        for _, command in self.commands.items():
            sub_parser = sub_parsers.add_parser(command.name, help=command.summary)
            for arg in command.args:
                self._factories[arg.arg_type](arg, sub_parser)
            sub_parser.set_defaults(__blnt__=command.name)
        parsed = vars(parser.parse_args())
        if '__blnt__' not in parsed:
            Console().error('Use the -h option to see CLI usage')
            sys.exit(1)
        cmd = parsed.pop('__blnt__')
        func = self.commands[cmd].func
        parsed['blnt'] = self.blnt
        parsed['context'] = self.blnt.context
        if inspect.iscoroutinefunction(func):
            loop = asyncio.get_event_loop()
            loop.run_until_complete(async_invoke(func, **parsed))
        else:
            invoke(func, **parsed)

    @staticmethod
    def _create_parser_arg(arg: Argument, *, optional: bool = False, use_flag: bool = False,
                           action: str = None):
        args = []
        kwargs = {}
        if optional:
            args.append(f'--{arg.name}')
        else:
            args.append(arg.name)
        if use_flag and arg.flag is not None:
            args.append(f'-{arg.flag}')
        if arg.summary is not None:
            kwargs['help'] = arg.summary
        if action is not None:
            kwargs['action'] = action
        if arg.value_type is not None:
            kwargs['type'] = arg.value_type
        return args, kwargs

    @staticmethod
    def _create_argument(arg: Argument, parser: ArgumentParser):
        args, kwargs = Parser._create_parser_arg(arg, optional=False, use_flag=False, action=None)
        parser.add_argument(*args, **kwargs)

    @staticmethod
    def _create_option(arg: Argument, parser: ArgumentParser):
        args, kwargs = Parser._create_parser_arg(arg, optional=True, use_flag=True, action=None)
        parser.add_argument(*args, **kwargs)

    @staticmethod
    def _create_flag(arg: Argument, parser: ArgumentParser):
        args, kwargs = Parser._create_parser_arg(arg, optional=True, use_flag=True, action='store_true')
        parser.add_argument(*args, **kwargs)



import asyncio
from json import JSONDecodeError
from typing import Dict, List, Union

import aiohttp
from aiohttp import web as aio_web
from aiohttp.web_request import Request

from bolinette import blnt, web
from bolinette.exceptions import APIError
from bolinette.utils.functions import async_invoke


class BolinetteSockets:
    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context
        self._topics: Dict[str, 'web.Topic'] = {}
        self._channels: Dict[str, List['web.TopicChannel']] = {}
        self._socket_sessions: Dict[str, aio_web.WebSocketResponse] = {}
        self._anon_socket_sessions: List[aio_web.WebSocketResponse] = []

    def add_topic(self, name: str, topic: 'web.Topic'):
        self._topics[name] = topic

    def topic(self, name: str) -> 'web.Topic':
        return self._topics.get(name)

    def add_channel(self, topic: str, channel: 'web.TopicChannel'):
        if topic not in self._channels:
            self._channels[topic] = []
        self._channels[topic].append(channel)

    def channels(self, topic: str) -> List['web.TopicChannel']:
        return self._channels.get(topic) or []
    
    async def send_message(self, topic: str, channels: Union[str, List[str]], message):
        socket_topic = self.topic(topic)
        if socket_topic is None:
            return
        if not isinstance(channels, list):
            channels = [channels]
        pending_tasks = []
        for channel in channels:
            subscriptions = socket_topic.subscriptions(channel)
            if subscriptions is not None:
                for subscription in subscriptions:
                    if subscription.closed:
                        continue
                    pending_tasks.append(subscription.send_json({
                       'topic': topic,
                       'channel': channel,
                       'message': message
                    }))
        await asyncio.gather(*pending_tasks)

    def add_socket_session(self, username: str, session: aio_web.WebSocketResponse):
        self._socket_sessions[username] = session

    def delete_socket_session(self, username: str):
        del self._socket_sessions[username]

    def add_anon_socket_session(self, session: aio_web.WebSocketResponse):
        self._anon_socket_sessions.append(session)

    def delete_anon_socket_session(self, session: aio_web.WebSocketResponse):
        self._anon_socket_sessions.remove(session)

    def init_socket_handler(self):
        self.context.app.add_routes([aio_web.get('/ws', SocketHandler().__call__)])


class SocketHandler:
    def __init__(self):
        pass

    async def __call__(self, request: Request):
        context: blnt.BolinetteContext = request.app['blnt']
        user_service = context.service('user')
        socket = aio_web.WebSocketResponse()
        await socket.prepare(request)

        current_user = None
        identity = context.jwt.verify(request, optional=True, fresh=True)
        if identity is not None:
            current_user = await user_service.get_by_username(identity)
        if current_user is not None:
            context.sockets.add_socket_session(current_user.username, socket)
        else:
            context.sockets.add_anon_socket_session(socket)

        async for msg in socket:  # type: aiohttp.WSMessage
            try:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    payload = msg.json()
                    action = payload['action']
                    if action == 'close':
                        await socket.close()
                    elif action == 'ping':
                        await socket.send_str('pong')
                    else:
                        await self.process_topic_message(context, socket, payload, current_user)
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    context.logger.warning(f'ws connection closed with exception {socket.exception()})')
            except APIError as e:
                context.logger.error(e.message)
            except JSONDecodeError:
                context.logger.error('sockets.non_deserializable_payload')

        if current_user is not None:
            context.sockets.delete_socket_session(current_user.username)
        else:
            context.sockets.delete_anon_socket_session(socket)
        return socket

    async def _subscribe(self, topic: web.Topic, socket: aio_web.WebSocketResponse, payload, current_user):
        if await async_invoke(topic.validate_subscription, payload=payload, socket=socket, current_user=current_user):
            await topic.receive_subscription(payload['channel'], socket)

    async def process_topic_message(self, context: 'blnt.BolinetteContext',
                                    socket: aio_web.WebSocketResponse, payload, current_user):
        action = payload['action']
        topic = context.sockets.topic(payload['topic'])
        if topic is not None:
            async with blnt.Transaction(context):
                if action == 'subscribe':
                    await self._subscribe(topic, socket, payload, current_user)
                elif action == 'send':
                    channels = context.sockets.channels(payload['topic'])
                    for channel in channels:
                        if channel.re.match(payload['channel']):
                            await async_invoke(channel.function, topic, socket=socket, payload=payload,
                                               current_user=current_user)



from bolinette import core, bcrypt
from bolinette.decorators import service
from bolinette.defaults.services import FileService
from bolinette.exceptions import ForbiddenError, UnprocessableEntityError


@service('user')
class UserService(core.Service):
    @staticmethod
    def encrypt_password(password):
        return bcrypt.hash_password(password)

    @staticmethod
    def check_password(user, password):
        return bcrypt.check_password(user.password, password)

    async def _check_params(self, values: dict):
        if values.get('password'):
            values['password'] = UserService.encrypt_password(values['password'])

    async def get_by_username(self, username, *, safe=False):
        return await self.get_first_by('username', username, safe=safe)

    async def get_by_email(self, email, *, safe=False):
        return await self.get_first_by('email', email, safe=safe)

    async def create(self, values: dict):
        await self._check_params(values)
        return await super().create(values)

    async def update(self, entity, values: dict):
        await self._check_params(values)
        return await super().update(entity, values)

    async def patch(self, entity, values: dict):
        await self._check_params(values)
        return await super().patch(entity, values)

    async def has_role(self, user, role_name):
        return any(filter(lambda r: r.name == role_name, user.roles))

    async def add_role(self, user, role):
        if role.name == 'root':
            raise ForbiddenError('role.root.forbidden')
        if role in user.roles:
            raise UnprocessableEntityError(f'user.roles.exists:{user.username}:{role.name}')
        user.roles.append(role)

    async def remove_role(self, current_user, user, role):
        if role.name == 'root':
            raise ForbiddenError('role.root.forbidden')
        if (current_user.username == user.username
                and role.name == 'admin'
                and not await self.has_role(current_user, 'root')):
            raise ForbiddenError('role.admin.no_self_demotion')
        if role not in user.roles:
            raise UnprocessableEntityError(f'user.roles.not_found:{user.username}:{role.name}')
        user.roles.remove(role)

    async def save_profile_picture(self, user, request_file):
        file_service: FileService = self.context.service('file')
        old_picture = user.profile_picture
        user.profile_picture = await file_service.save_file(request_file)
        if old_picture:
            await file_service.delete(old_picture)
        return user



import os
import random
import string


def cwd():
    return os.getcwd()


def random_string(length):
    return ''.join(random.choices(string.ascii_letters + string.digits, k=length))


def mkdir(path):
    os.makedirs(path)


def exists(path):
    return os.path.exists(path)


def rename(path, new_path):
    os.rename(path, new_path)


def join(*args):
    return os.path.join(*args)


def dirname(path):
    return os.path.dirname(os.path.realpath(path))


def split(path):
    return os.path.split(path)


def rm(path):
    os.remove(path)


def rm_r(path):
    for root, dirs, files in os.walk(path, topdown=False):
        for name in files:
            os.remove(os.path.join(root, name))
        for name in dirs:
            os.rmdir(os.path.join(root, name))
    os.rmdir(path)



from typing import Dict, Callable, Iterator, Tuple

from bolinette import core, blnt, types


class Mixin:
    def __init__(self):
        self.__props__ = MixinProps(self)
        self._methods = {}
        for method_name, method in self.__props__.get_service_methods():
            self._methods[method_name] = method

    def __contains__(self, key):
        return key in self._methods

    def __getitem__(self, key):
        return self._methods[key]

    def columns(self) -> Dict[str, 'core.models.Column']:
        pass

    def relationships(self, model) -> Dict[str, 'types.defs.Relationship']:
        pass

    def payload(self, model):
        pass

    def response(self, model):
        pass


class MixinProps(blnt.Properties):
    def get_service_methods(self) -> Iterator[Tuple[str, 'MixinServiceMethod']]:
        return self._get_cls_attributes_of_type(type(self.parent), MixinServiceMethod)


class MixinServiceMethod:
    def __init__(self, name: str, func: Callable):
        self.name = name
        self.func = func



from abc import ABC, abstractmethod


class DatabaseEngine(ABC):
    def __init__(self, relational: bool):
        self.relational = relational

    @abstractmethod
    async def open_transaction(self):
        pass

    @abstractmethod
    async def close_transaction(self):
        pass

    @abstractmethod
    async def rollback_transaction(self):
        pass

    @abstractmethod
    async def create_all(self):
        pass

    @abstractmethod
    async def drop_all(self):
        pass



from abc import ABC, abstractmethod
from typing import Any, Dict, Callable, Awaitable

from aiohttp.web_request import Request
from aiohttp.web_response import Response

from bolinette import blnt


class Middleware:
    __blnt__: 'MiddlewareMetadata' = None

    def __init__(self, context: 'blnt.BolinetteContext'):
        self.system_priority = 1
        self.context = context
        self.options = {}
        self.params = MiddlewareParams()

    def define_options(self) -> Dict[str, 'MiddlewareParam']:
        return {}

    async def handle(self, request: Request, params: Dict[str, Any],
                     next_func: Callable[[Request, Dict[str, Any]], Awaitable[Response]]):
        return await next_func(request, params)

    def __repr__(self):
        return f'<Middleware {self.__blnt__.name}>'


class MiddlewareMetadata:
    def __init__(self, name: str, priority: int, auto_load: bool, loadable: bool):
        self.name = name
        self.priority = priority
        self.auto_load = auto_load
        self.loadable = loadable


class InternalMiddleware(Middleware):
    def __init__(self, context: 'blnt.BolinetteContext'):
        super().__init__(context)
        self.system_priority = 0


class MiddlewareParams:
    @staticmethod
    def bool(required=False, default=False):
        return BooleanParam(required, default)

    @staticmethod
    def string(required=False, default=None):
        return StringParam(required, default)

    @staticmethod
    def int(required=False, default=None):
        return IntParam(required, default)

    @staticmethod
    def float(required=False, default=None):
        return FloatParam(required, default)

    @staticmethod
    def list(element: 'MiddlewareParam', required=False, default=None):
        if default is None:
            default = []
        return ListParam(element, required, default)


class MiddlewareParam(ABC):
    def __init__(self, required=False, default=None):
        self.required = required
        self.default = default

    @abstractmethod
    def validate(self, value: str):
        pass


class BooleanParam(MiddlewareParam):
    def validate(self, value: str):
        return True

    def __repr__(self) -> str:
        return 'bool'


class StringParam(MiddlewareParam):
    def validate(self, value: str):
        return value

    def __repr__(self) -> str:
        return 'str'


class IntParam(MiddlewareParam):
    def validate(self, value: str):
        return int(value)

    def __repr__(self) -> str:
        return 'int'


class FloatParam(MiddlewareParam):
    def validate(self, value: str):
        return float(value)

    def __repr__(self) -> str:
        return 'float'


class ListParam(MiddlewareParam):
    def __init__(self, element: MiddlewareParam, required=False, default=None):
        super().__init__(required, default)
        self.element = element

    def validate(self, value: str):
        return [self.element.validate(val) for val in value.split(',')]

    def __repr__(self) -> str:
        return f'list[{repr(self.element)}]'



import traceback

from sqlalchemy.exc import SQLAlchemyError

from bolinette import blnt
from bolinette.exceptions import APIError, APIErrors, InternalError


class Transaction:
    def __init__(self, context: 'blnt.BolinetteContext', *,
                 print_error: bool = True, propagate_error: bool = True, raise_internal: bool = True):
        self.context = context
        self.print_error = print_error
        self.propagate_error = propagate_error
        self.raise_internal = raise_internal

    async def __aenter__(self):
        await self.context.db.open_transaction()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            await self.context.db.rollback_transaction()
            if not issubclass(exc_type, (APIError, APIErrors)):
                if self.print_error:
                    self.context.logger.error(str(exc_val))
                    traceback.print_tb(exc_tb)
                if self.propagate_error and self.raise_internal:
                    if self.context.env['debug']:
                        raise InternalError([str(exc_val)] + traceback.format_list(traceback.extract_tb(exc_tb)))
                    raise InternalError('internal.error')
                elif self.propagate_error:
                    raise exc_val
        else:
            try:
                await self.context.db.close_transaction()
            except SQLAlchemyError as err:
                await self.context.db.rollback_transaction()
                raise InternalError([f'global.internal_error:{err}'])



class APIError(Exception):
    def __init__(self, message, *, name="APIError"):
        super().__init__(name)
        self.message = message

    def __str__(self):
        return self.message

    def __repr__(self):
        return f'<APIError {self.message}>'


class APIErrors(Exception):
    def __init__(self):
        self.errors = []

    def append(self, error: APIError):
        self.errors.append(error)

    def __bool__(self):
        return len(self.errors) > 0

    def __repr__(self):
        return f'<APIErrors [{",".join([repr(err) for err in self.errors])}]>'


class InternalError(APIError):
    def __init__(self, message, *, name='InternalError'):
        super().__init__(message, name=name)


class NotFoundError(APIError):
    def __init__(self, message, *, name='NotFoundError'):
        super().__init__(message, name=name)


class ConflictError(APIError):
    def __init__(self, message, *, name='ConflictError'):
        super().__init__(message, name=name)


class BadRequestError(APIError):
    def __init__(self, message, *, name='BadRequestError'):
        super().__init__(message, name=name)


class UnprocessableEntityError(APIError):
    def __init__(self, message, *, name='UnprocessableEntityError'):
        super().__init__(message, name=name)


class ForbiddenError(APIError):
    def __init__(self, message, *, name='ForbiddenError'):
        super().__init__(message, name=name)


class UnauthorizedError(APIError):
    def __init__(self, message, *, name='UnauthorizedError'):
        super().__init__(message, name=name)


class EntityNotFoundError(NotFoundError):
    def __init__(self, model, key, value):
        super().__init__(f'entity.not_found:{model}:{key}:{value}', name='EntityNotFoundError')


class ParamMissingError(UnprocessableEntityError):
    def __init__(self, key):
        super().__init__(f'param.required:{key}', name='ParamMissingError')


class ParamNonNullableError(UnprocessableEntityError):
    def __init__(self, key):
        super().__init__(f'param.non_nullable:{key}', name='ParamMissingError')


class ParamConflictError(ConflictError):
    def __init__(self, key, value):
        super().__init__(f'param.conflict:{key}:{value}', name='ParamConflictError')


class InitError(Exception):
    def __init__(self, message: str = None, *, inner: Exception = None):
        self.message = message
        self.inner = inner

    def __str__(self):
        if self.message is None:
            return str(self.inner)
        return self.message



from typing import Type, TypeVar, Generic

T = TypeVar('T')


class InitProxy(Generic[T]):
    def __init__(self, proxy_cls: Type[T], **kwargs):
        self._proxy_cls = proxy_cls
        self._kwargs = kwargs

    def instantiate(self, **kwargs) -> T:
        return self._proxy_cls(**(self._kwargs | kwargs))

    def of_type(self, proxy_cls: Type[T]) -> bool:
        return self._proxy_cls == proxy_cls

    def __repr__(self):
        return f'<InitProxy #{self.__hash__()}: {repr(self._proxy_cls)}>'



from typing import List, Dict, Any

from bolinette import blnt, core
from bolinette.blnt.objects import Pagination, PaginationParams, OrderByParams
from bolinette.blnt.database.queries import BaseQuery, RelationalQueryBuilder, CollectionQueryBuilder
from bolinette.exceptions import ParamConflictError, APIErrors, ParamNonNullableError
from bolinette.utils.functions import setattr_, getattr_


class Repository:
    def __init__(self, name: str, model: 'core.Model', context: 'blnt.BolinetteContext'):
        self.name = name
        self.model = model
        self.context = context
        self._query_builder = self._get_query_builder(model, context)

    @staticmethod
    def _get_query_builder(model: 'core.Model', context: 'blnt.BolinetteContext'):
        db = context.db[model.__blnt__.database]
        if db.relational:
            return RelationalQueryBuilder(model, context)
        return CollectionQueryBuilder(model, context)

    def __repr__(self):
        return f'<Repository {self.name}>'

    def query(self):
        return self._query_builder.query()

    async def get_all(self, pagination: PaginationParams = None, order_by: List[OrderByParams] = None):
        if order_by is None:
            order_by = []
        query = self.query()
        if len(order_by) > 0:
            for order in order_by:
                query = query.order_by(order.column, desc=not order.ascending)
        if pagination is not None:
            return await self._paginate(query, pagination)
        return await query.all()

    async def get(self, identifier):
        return await self.query().get_by_id(identifier)

    async def get_by(self, key: str, value):
        return await self.query().filter_by(**{key: value}).all()

    async def get_first_by(self, key: str, value):
        return await self.query().filter_by(**{key: value}).first()

    async def create(self, values: Dict[str, Any]):
        filtered = await self._validate_model(values)
        entity = await self._query_builder.insert_entity(filtered)
        return entity

    async def update(self, entity, values: Dict[str, Any]):
        await self._map_model(entity, values)
        return await self._query_builder.update_entity(entity)

    async def patch(self, entity, values: Dict[str, Any]):
        await self._map_model(entity, values, patch=True)
        return await self._query_builder.update_entity(entity)

    async def delete(self, entity):
        return await self._query_builder.delete_entity(entity)

    @staticmethod
    async def _paginate(query: BaseQuery, pagination: PaginationParams):
        page = pagination.page
        per_page = pagination.per_page
        total = await query.count()
        items = await query.offset(page * per_page).limit(per_page).all()
        return Pagination(items, page, per_page, total)

    async def _validate_model(self, values: Dict[str, Any]):
        api_errors = APIErrors()
        ignore_keys = []
        for _, relationship in self.model.__props__.get_relationships():
            key = relationship.name
            if key in values and values[key] is not None:
                ignore_keys.append(relationship.foreign_key.name)
        for _, column in self.model.__props__.get_columns():
            key = column.name
            if column.primary_key or key in ignore_keys:
                continue
            if key in values:
                value = values.get(key)
                if value is None and not column.nullable:
                    api_errors.append(ParamNonNullableError(key))
                if column.unique:
                    if await self.get_first_by(column.name, value) is not None:
                        api_errors.append(ParamConflictError(key, value))
            elif not column.nullable:
                api_errors.append(ParamNonNullableError(key))
        if api_errors:
            raise api_errors
        return values

    async def _map_model(self, entity, values: Dict[str, Any], patch=False):
        api_errors = APIErrors()
        for _, column in self.model.__props__.get_columns():
            key = column.name
            if column.primary_key or (key not in values and patch):
                continue
            original = getattr_(entity, key, None)
            new = values.get(key, None)
            if original == new:
                continue
            if new is None and not column.nullable:
                api_errors.append(ParamNonNullableError(key))
            if column.unique and new is not None:
                if await self.get_first_by(column.name, new) is not None:
                    api_errors.append(ParamConflictError(key, new))
            setattr_(entity, key, new)
        if api_errors:
            raise api_errors



import bcrypt as py_bcrypt


class Bcrypt:
    def hash_password(self, password: str):
        return py_bcrypt.hashpw(password.encode(), py_bcrypt.gensalt()).decode()

    def check_password(self, hashed: str, password: str):
        return py_bcrypt.checkpw(password.encode(), hashed.encode())


bcrypt = Bcrypt()
